#! /usr/bin/python

from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from cgi import parse_qs
from httplib import HTTPConnection
from lxml.etree import Element, SubElement, fromstring, tostring, parse
from ConfigParser import ConfigParser
from sys import argv
from syslog import syslog # only Unix
from urllib import urlopen
from urlparse import urlparse
from xml.sax.saxutils import unescape

####
## This gateway handles sruUpdates of a Meresco harvester (or any other server that sends sruUpdates).
## There are two kinds of updates: adding jel-codes to eo-records and adding ddi-documents to eo-records
## The sruUpdates have to be processed by the gateway to make them suitable as sruUpdates that can be sent to a
## a (Meresco) writer
####

diag_ns = 'http://www.loc.gov/zing/srw/diagnostic/'
meta_ns = 'http://meresco.com/namespace/harvester/meta'
mods_ns = 'http://www.loc.gov/mods/v3'
srw_ns  = 'http://www.loc.gov/zing/srw/'
ucp_ns  = 'info:lc/xmlns/update-v1'
ns = {'ddi' :'http://www.icpsr.umich.edu/DDI',
      'ddi3': 'ddi:instance:3_1',
      'r'   : 'ddi:reusable:3_1',
      's'   : 'ddi:studyunit:3_1',
      'diag': diag_ns,
      'didl': 'urn:mpeg:mpeg21:2002:02-DIDL-NS',
      'dii' : 'urn:mpeg:mpeg21:2002:01-DII-NS',
      'meta': meta_ns,
      'mods': mods_ns,
      'srw' : srw_ns,
      'ucp' : ucp_ns}

class SruUpdateHandler(BaseHTTPRequestHandler):
    def doRequest(self, query, xml, part="", target=""):
        """Preparing and sending a request to a remote writer.

        A sruUpdate request from a harvester can result in more than one sruUpdate requests to a writer.
        For each request this function has to be called.
        query: used in sru request that searches for the publication that must be enriched
        xml: object with the enrichment info as supplied by the sender of the sruUpdate request (harvester).
        part: name of part that must be added to the record
        """
        if not target: target = part
        baseurl = config.get('reader', target)
        url = baseurl + "?version=1.1&operation=searchRetrieve&x-recordSchema=meta&query=" + query
        #print "url: " + url
        try:
            sruResponse = parse(urlopen(url))
        except IOError, (errno, strerror):
            self._status = "fail"
            self._details += " baseurl=%s errno=%s strerror=%s" % (baseurl, errno, strerror)
        else:
            #print tostring(sruResponse, pretty_print=True)
            if sruResponse.xpath("//srw:numberOfRecords/text()", ns)[0] == '0':
                ## no publications found ##
                self._status = "fail"                
                self._details += " reason=record not found in remote database"
            else:
                ## publication found. FIXME what happens when # hits > 1?
                ## get info from meta-part ##
                meta = sruResponse.xpath("//meta:meta", ns)[0]
                recordIdentifier = meta.xpath("meta:record/meta:id/text()", ns)[0]
                repository       = meta.xpath("meta:repository", ns)[0]
                collection   = repository.xpath("meta:collection/text()", ns)[0]
                repositoryId = repository.xpath("meta:id/text()", ns)[0]
                self._details += " uploadId=" + repositoryId + ":" + recordIdentifier

                ## new Meta object ##
                meta = Meta(meta, part).get()
                #print meta

                ## creating new Document object ##
                if part == "ddi" or part == "ddi3":
                    document = Document(meta, tostring(xml), part=part).get(method="Element")
                elif part == "jel":
                    jels = xml.xpath\
                        ("//didl:DIDL/didl:Item/didl:Component/didl:Resource/mods:mods/mods:classification[@authority='info:eu-repo/classification/jel']/text()", ns)
                    if jels:
                        mods = Mods(jels).get()
                        #print mods
                        document = Document(meta, mods, part="jel").get(method="Element")
                    else:
                        self._status = "fail"
                        self._details += " reason=no jel-codes supplied"

                ## when there is a new Document, it can be sent to the appropriate writer ##
                if document:
                    recordUpdate = UpdateRequest("replace", repositoryId, recordIdentifier, document)
                    #print recordUpdate.get(xml_declaration=True)

                    ## determine port of writer; the collection determines the port
                    port = config.get('default_writer', 'port')
                    port = config.get(collection, 'port')
                    self._details += " wport=" + port
                    
                    self._status, d = recordUpdate.send(port)
                    if d: self._details += " diagnostic=" + d

        finally:
            updateResponse = UpdateResponse(self._status, details=self._details, message=self._message).get(xml_declaration=True)
            #print updateResponse
            #print self._status, self._details
            if self._status == "fail":
                syslog("enrichment failure" + self._details)
            self.wfile.write(updateResponse)

    def getIdentifierFromRelPubl(self, relPubl):
        """Identifiers are filtered from the ddi field 'relPub'.

        The variable 'permalinks' is global and is defined in main()
        In 'permalinks' the forms of the permalinks are specified.
        A permalink form is used to locate an identifier.
        The variable 'permalinks' is a list of (permalink form, target) pairs.
        The function returns a list of (identifier, target) pairs.
        """
        identifiers = []
        for permalink in permalinks:
            loc = relPubl.find(permalink[0])
            if loc < 0: continue
            s = relPubl[loc + len(permalink[0]):]
            locquote = s.find('\"')
            if locquote < 0: locquote = s.find("\'")
            if locquote > 0: identifier = s[:locquote]
            else: identifier = s
            identifier = identifier.strip()
            target = permalink[1]
            identifiers.append((identifier, target))
            #print "identifier = " + identifier + "|"
        return identifiers

    def do_POST(self):
        self._status = "fail"
        self._details = ""
        self._message = ""
        try:
            updateRequest = self.rfile.read(int(self.headers['Content-Length']))
            self.send_response(200)
            self.send_header("Content-Type", "text/xml, charset=utf-8")
            self.end_headers()
  
            xml = fromstring(unescape(updateRequest))
            #print tostring(xml)

            codeBook = xml.xpath("//ddi:codeBook", ns)
            if codeBook:
                codeBook = codeBook[0]
                #print tostring(codeBook, pretty_print=True)
                relPubls = codeBook.xpath("//ddi:codeBook/ddi:stdyDscr/ddi:othrStdyMat/ddi:relPubl/text()", ns)
                for relPubl in relPubls:
                    #print "relPubl: " + relPubl
                    identifiers = self.getIdentifierFromRelPubl(relPubl)
                    if not identifiers: continue
                    for idpair in identifiers:
                        identifier = idpair[0]
                        target = idpair[1]
                        self._details += " type=ddi pubid=" + identifier
                        # we zoeken alleen op de record identifier uit het pubid
                        query = "meta.record.id%20exact%20%22" + identifier.split(':', 1)[1] + "%22"
                        self.doRequest(query, codeBook, part="ddi", target=target)
                        # initialising for next round
                        self._status = "fail"
                        self._details = ""
                        self._message = ""
           
            DIDL = xml.xpath("//didl:DIDL", ns)
            if DIDL:
                DIDL = DIDL[0]
                #print tostring(DIDL, pretty_print=True)
                docid = DIDL.xpath("//didl:DIDL/didl:Descriptor/didl:Statement/dii:Identifier/text()", ns)[0]
                self._details += " type=jel docid=" + docid
                #print docid
                query = "norm.docid%20exact%20%22" + docid + "%22"
                self.doRequest(query, DIDL, part="jel" )
                
            DDI = xml.xpath("//ddi3:DDIInstance", ns)
            if DDI:
                DDI = DDI[0]
                url = DDI.xpath("//s:StudyUnit/r:OtherMaterial/r:ExternalURLReference/text()", ns)[0]
                identifier = parse_qs(urlparse(url).query)['identifier'][0]
                self._details += " type=ddi3 pubid=" + identifier
                query = "meta.record.id%20exact%20%22" + identifier + "%22"
                self.doRequest(query, DDI, part="ddi3")
        except:
            pass

class Tree(object):
    def __init__(self):
        self.root = None

    def get(self, method="xml", xml_declaration=False):
        if method == "Element":
            return self.root
        elif method == "xml":
            return tostring(self.root, xml_declaration=xml_declaration, encoding="utf-8")

class Mods(Tree):
    def __init__(self, jels):
        self.ns = {'xmlns:mods': mods_ns}
        self.root = Element('mods:mods', self.ns, version="3.2")
        for jel in jels:
            SubElement(self.root, 'mods:classification', authority="info:eu-repo/classification/jel").text = jel

class Meta(Tree):
    """This is a sort of factory class for adding information about the part that is added to the document.

    To a meta part a 'parts' element and a sub element 'part' is added: meta/parts/part=partName.
    The value of the sub element is the name of the part that is added to the document.
    If the 'parts' element already exists, the 'part' sub element is added.
    If the latter already exists, nothing is done with the meta part."""
    def __init__(self, meta, partName):
        self.ns = {'xmlns': meta_ns}
        parts = meta.xpath("meta:parts/meta:part", ns)
        if parts:
            partNameExists = False
            for part in parts:
               if part.text == partName:
                    partNameExists = True
                    break
            if not partNameExists:
                part = Element('part', self.ns)
                part.text = partName
                parts[0].getparent().append(part)
        else:
            parts = Element('parts', self.ns)
            SubElement(parts, 'part').text = partName
            meta.append(parts)
        self.root = meta

class Document(Tree):
    def __init__(self, meta, record, part=""):
        self.root = Element('document', xmlns="http://meresco.com/namespace/harvester/document")
        SubElement(self.root, 'part', name="meta").text = meta
        SubElement(self.root, 'part', name=part).text = record 

class UpdateResponse(Tree):
    def __init__(self, status, details="", message=""):
        self.ns = {"xmlns:srw": srw_ns, "xmlns:ucp": ucp_ns}
        self.root = Element('srw:updateResponse', self.ns)
        SubElement(self.root, 'srw:version').text = "1.0"
        SubElement(self.root, 'ucp:operationStatus').text = status
        if status == "fail":
            self.ns = {"xmlns:diag": diag_ns}
            self.diag = SubElement(self.root, 'diag:diagnostic', self.ns)
            SubElement(self.diag, 'diag:uri').text = "info:srw/diagnostic/12/1"
            SubElement(self.diag, 'diag:details').text = details
            SubElement(self.diag, 'diag:message').text = "Invalid component: record rejected"

class UpdateRequest(Tree):
    def __init__(self, action, repository, recordIdentifier, document):
        self.ns = {"xmlns:srw": srw_ns, "xmlns:ucp": ucp_ns}
        self.root = Element('srw:updateRequest', self.ns)
        SubElement(self.root, 'srw:version').text = "1.0"
        self.action = SubElement(self.root, 'ucp:action')
        if action == "replace" : self.action.text = "info:srw/action/1/replace"
        elif action == "delete": self.action.text = "info:srw/action/1/delete"
        SubElement(self.root, 'ucp:recordIdentifier').text = repository + ":" + recordIdentifier
        if action == "replace":
            self.record = SubElement(self.root, 'srw:record')
            SubElement(self.record, 'srw:recordPacking').text = 'xml'
            SubElement(self.record, 'srw:recordSchema').text  = 'document'
            SubElement(self.record, 'srw:recordData').append(document)

    def send(self, port):
        # FIXME HTTPExceptions afvangen
        self.request = self.get(xml_declaration=True)
        #print self.request
        ## writer waarnaar de enrichment wordt gestuurd ##
        connection = HTTPConnection('localhost', port)
        connection.putrequest("POST", "/update")
        connection.putheader("Content-Type", "text/xml; charset=\"utf-8\"")
        connection.putheader("Content-Length", str(len(self.request)))
        connection.endheaders()
        connection.send(self.request)

        response = connection.getresponse()
        status = response.status
        content = fromstring(response.read())
        #print tostring(content)
        if status == 200:
            if content.xpath("//ucp:operationStatus/text()", ns)[0] == "success":
                return "success", ""
            else:
                return "fail", content.xpath("//diag:diagnostic/diag:details/text()", ns)[0]
        else:
            return "fail", "HTTP problem: %s" %status

def main():
    args=argv[1:]
    if len(args) != 1: 
        print "supply one argument for the config file"
        exit()
    else:
        # FIXME mooi is anders, maar ik weet niet beter.
        # BaseHTTPRequestHandler heeft een instance variable server die gebruikt kan worden,
        # maar ik weet niet hoe.
        global config
        config = ConfigParser()
        config.read(args[0])
        global permalinks
        permalinks = []
        cf = config.get('dvn', 'permalinks')
        print cf
        for element in cf.split():
            parts = element.split("^")
            permalinks.append((parts[0],parts[1]))
    try:
        port = int(config.get('server', 'port'))
        server = HTTPServer(('', port), SruUpdateHandler)
        print 'started httpserver...'
        syslog("started enrichment server at "    + config.get('server', 'port'))
        syslog("enrichment port default writer: " + config.get('default_writer', 'port'))
        syslog("enrichment port uvt: "            + config.get('uvt', 'port'))
        syslog("enrichment port eo repo: "        + config.get('eo-repo', 'port'))
        syslog("enrichment port repec: "          + config.get('repec', 'port'))
        syslog("enrichment port evs: "            + config.get('evs', 'port'))
        server.serve_forever()
    except KeyboardInterrupt:
        print '^C received, shutting down server'
        server.socket.close()

if __name__ == '__main__':
    main()
